#
# $Copyright: Copyright (c) 2009 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $RCSfile: DS4KRemoteMirror.pm,v $ $Revision: 1.1.2.5 $
#

package DS4KRemoteMirror;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);
use ag_i18n_inc;

@EXPORT = qw(
);
$VERSION = '1.00';

require 5.000;

# Common variables concerning OS / logical volume management
use POSIX qw(uname);
(my $os_name) = uname;

# ESX specific variables
my $java_home="/opt/VRTSjre/jre1.5/bin";
my $vcs_home=$VCSHOME;
my $JAVA = "$java_home/java";
my $JAVA_CLASSPATH=join (':', <$VCSHOME/vcsvi-sdk/jars/*.jar>);
my $VMWARE_API_CMD = "$JAVA -cp \"$JAVA_CLASSPATH\" com.vmf.VirtualEntityComDriver";
my $cred_file="/etc/VRTSvcs/conf/config/\.esxlocalhostWSCred";


sub new {
	my($self) = shift;
        my($resname) = shift;
        my($smclihome) = shift;
	
	my $i=0;
	my ($numLocalStorage) = shift;
	my @LocalStorageNames;
	for($i=0; $i<$numLocalStorage; $i++){
  		@LocalStorageNames[$i] = shift;
  	}

        my($numLogicalDrives) = shift;
	my @LogicalDriveNames;

	my (%ld_state) = ();
	my (%ld_role) = ();
	my (%ld_id) = ();
	
	# populate our LogicalDriveNames array
	for($i=0; $i<$numLogicalDrives; $i++){
    		$LogicalDriveNames[$i] = shift;
		$ld_state{@LogicalDriveNames[$i]} = "";
		$ld_role{@LogicalDriveNames[$i]} = "";
		$ld_id{@LogicalDriveNames[$i]} = "";
	}
        
	my($storagename)="NOSTORAGE";
	for($i=0; $i<$numLocalStorage; $i++){
		if (PingTest($LocalStorageNames[$i])) {
        		VCSAG_LOG_MSG("N", "Local Storage Sub System $LocalStorageNames[$i] is unreachable.", 1, $LocalStorageNames[$i] );
		} else {
			$storagename = $LocalStorageNames[$i];
			last;
		}
	}


	$self = {
		"resname" => $resname,
		"smclihome" => $smclihome,
		"storage" => $storagename,
		"numLDs" => $numLogicalDrives,
		"smcli" => '',
		"effrepstate" => '', # Effective Replication State
		"smcout" => '',
		"err" => '',
		"platform" => $os_name,
		"effrole" => '' # Effective Role of all Logical Drives, 
				# it could be primary, secondary or mixed.
				# mixed means, some logcal drive's role is primary and
				# some are secondary.
	};

	$self->{LocalStorageNames} = \@LocalStorageNames;
	$self->{LogicalDriveNames} = \@LogicalDriveNames;
	$self->{LD_STATE} = \%ld_state;
	$self->{LD_ROLE} = \%ld_role;
	$self->{LD_ID} = \%ld_id;

	if ($os_name eq "Windows NT") {
		$self->{'smcli'} = "$self->{'smclihome'}" . "\\SMcli ";
	} else {
		$self->{'smcli'} = "$self->{'smclihome'}" . "/SMcli ";
   	}

	if ($self->{'storage'}) {
		$self->{'smcli'} = $self->{'smcli'} . "$self->{'storage'}" . " -S -c ";
	}

	my $class = ref($self) || $self;
	bless $self;
	return $self;
}

sub DESTROY {}

sub get_state {
	my $self = shift;

	my $ret = 0;
	my $i = 0;

	for($i=0; $i < $self->{'numLDs'}; $i++){
		my $ldrive = $self->{LogicalDriveNames}[$i];

		($self->{LD_ROLE}{$ldrive}, $self->{LD_STATE}{$ldrive}) = $self->get_logical_drive_role_and_state($ldrive);
		VCSAG_LOG_MSG("I", "Logical Drive [$ldrive] is in state: $self->{LD_STATE}{$ldrive} and role: $self->{LD_ROLE}{$ldrive}", 224, $self->{LD_STATE}{$ldrive}, $self->{LD_ROLE}{$ldrive});
	}

	$self->{'effrepstate'} = $self->get_effective_replication_state();

	$ret = $self->is_link_ok();

	if ($ret) {
		VCSAG_LOG_MSG("W", "Health of Remote Mirror is not optimal, hence setting Effective Replication state to Unsynchronized", 225);
		$self->{'effrepstate'} = "Unsynchronized";
	}

	$self->{'effrole'} = $self->get_effective_role();

	return $ret;
}

sub get_effective_role {
	my ($self) = shift;
	my $eff_role = "";
	my $role = "";
	my $i;

	for($i=0; $i < $self->{'numLDs'}; $i++){
		my $ldrive = $self->{LogicalDriveNames}[$i];

		$role = $self->{LD_ROLE}{$ldrive};

		if ( ($eff_role ne "") && ($role ne $eff_role) ) {
			$eff_role = "mixed";
			last;
		}

		$eff_role = $role;
	}

	return $eff_role;
}


sub get_effective_replication_state {
	my ($self) = shift;
	my $eff_rep_state = "";
	my $i;

	for($i=0; $i < $self->{'numLDs'}; $i++){
		my $ldrive = $self->{LogicalDriveNames}[$i];
		my $state = $self->{LD_STATE}{$ldrive};

		$eff_rep_state = $state;

		if ( $state eq "Synchronized" ) {
			next;
		} else {
			last;
		}
	}

	return $eff_rep_state;
}

sub is_link_ok {
	my ($self) = shift;
	my $ret;
	my $script_cmd = {};

        $script_cmd->{'cmd'} = "\"show storagesubsystem healthStatus;\"";
        $script_cmd->{'logcmd'} = "$self->{'smcli'} show storagesubsystem healthStatus;";

        if ($self->execute_smcli($script_cmd)) {
                VCSAG_LOG_MSG("E", "healthstatus check of storagesubsystem request failed with message:$self->{'smcout'}", 121,"$self->{'smcout'}");
                return 1;
        }

        foreach my $line (split('\n', $self->{'smcout'})) {
		if ( $line =~ /Unable to contact remote storage subsystem/) {
			return 2;
		}

		if ( $line =~ /Mirror Communication Error - Unable to Contact Fabric/) {
			return 3;
		}

		if ( $line =~ /Secondary storage subsystem: Unable to contact remote storage subsystem/) {
			return 4;
		}

		if ( $line =~ /Primary storage subsystem: Unable to contact remote storage subsystem/) {
			return 5;
		}

		if ( $line =~ /Mirror Communication Error - Unable to Contact Storage Subsystem/) {
			return 6;
		}
        }

	return 0;
}


sub get_role_and_state_if_ldrive_names_equal {
	my ($self) = shift;
        my $ldrive = shift;
        my $role   = "";
        my $role1   = "";
        my $role2   = "";
        my $rep_state = "";
        my $script_cmd = {};

        $script_cmd->{'cmd'} = "\"show alllogicaldrives;\"";
        $script_cmd->{'logcmd'} = "$self->{'smcli'} show alllogicaldrives;";

        if ($self->execute_smcli($script_cmd)) {
                VCSAG_LOG_MSG("E", "All logical drives info request failed with message:$self->{'smcout'}", 111,"$self->{'smcout'}");
                return ($role, $rep_state);
        }
	
	my $search_flag = 0;
	my $search_str = "Mirrored pair: $ldrive and $ldrive";
	my $ld1 = "";
	my $ld2 = "";
	my $ldid1 = "";
	my $ldid2 = "";

	$ldid1 = $self->{LD_ID}{$ldrive};
	foreach my $line (split('\n', $self->{'smcout'})) {
		if ($line =~ /$search_str/) {
			$search_flag = 1;
			next;
		}
		
		if ($line =~ /Primary logical drive/) {
			if ( $search_flag == 0 ) {
				next;
			}

			$role1 = "primary";
                        $ld1 = $ldrive;
		}

		if ($line =~ /Secondary logical drive/) {
                        if ( $search_flag == 0 ) {
                                next;
                        }

                        $role2 = "secondary";
                        $ld2 = $ldrive;
                }
	
		if ($line =~ /Logical Drive ID/) {
                        if ( $search_flag == 0 ) {
                                next;
                        }

                        $ldid2 = ( $line =~ /[\S]+/g )[-1];
			if (($ldid1 eq $ldid2) && ($ld1 ne "") && ($ld2 eq "")) {
				$role = $role1;
			} else {
				$role = $role2;
			}
                }

		if ($line =~ /Role/) {
                        if ( $search_flag == 0 ) {
                                next;
                        }

			if ( $ld1 eq "" ) {
				$ld1 = $ldrive;
                        	$role1 = ( $line =~ /[\S]+/g )[-1];
				$role1 =~ tr/A-Z/a-z/;
			} else {
				$ld2 = $ldrive;
                        	$role2 = ( $line =~ /[\S]+/g )[-1];
				$role2 =~ tr/A-Z/a-z/;
			}
                }

		if ($line =~ /Mirror status/) {
			if ( $search_flag == 0 ) {
                                next;
                        }
                        $rep_state = ( $line =~ /[\S]+/g )[-1];
		}
	}	
	
	return ($role, $rep_state);
}

sub get_logical_drive_role_and_state {
	my ($self) = shift;
	my $ldrive = shift;
	my $role   = "";
	my $rep_state = "";
	my $script_cmd = {};

	$script_cmd->{'cmd'} = "\"show remoteMirror localLogicalDrive [\\\"$ldrive\\\"] synchronizationProgress;\"";
	$script_cmd->{'logcmd'} = "$self->{'smcli'} show remoteMirror localLogicalDrive [$ldrive] synchronizationProgress;";

	if ($self->execute_smcli($script_cmd)) {
		VCSAG_LOG_MSG("E", "All logical drives info request failed with message:$self->{'smcout'}", 111,"$self->{'smcout'}");
		return ($role, $rep_state);
	}
	
	foreach my $line (split('\n', $self->{'smcout'})) {
		if ($line =~ /Mirror pair/) {
			$line =~ s/[ ():\/]/ /g;
			my (undef, undef, $r1, $ld1, $r2, $ld2, $s) = split ('\s+', $line);
			if ( $ld1 eq $ld2) {
				$self->{LD_ID}{$ldrive} = $self->get_logical_drive_ID($ldrive);
				($role, $rep_state) = $self->get_role_and_state_if_ldrive_names_equal($ldrive);	
				return ($role, $rep_state);
			}

			if ( "$ld1" eq "$ldrive") {
				$role = $r1;
			}

			if ( "$ld2" eq "$ldrive") {
				$role = $r2;
			}

			$rep_state = $s;
			last;
		}
	}

	return ($role, $rep_state);	
}

sub get_logical_drive_ID {
	my ($self) = shift;
	my $ldrive = shift;
	my $LDID = "";
	my $script_cmd = {};

	$script_cmd->{'cmd'} = "\"show logicalDrive [\\\"$ldrive\\\"];\"";
	$script_cmd->{'logcmd'} = "$self->{'smcli'} show logicalDrive [$ldrive];";
	
	if ($self->execute_smcli($script_cmd)) {
		VCSAG_LOG_MSG("E", "LogicalDrive info request failed with message:$self->{'smcout'}", 111, "$self->{'smcout'}");
		return $LDID;
	}
	
	foreach my $line (split('\n', $self->{'smcout'})) {
		if ($line =~ /Logical Drive ID/) {
			my (@ldidline) = split('\s+', $line);
			$LDID = $ldidline[4];
			last;
		}
	}
	
	return $LDID;
}

sub failover_required {
	my ($self) = shift;

	VCSAG_LOG_MSG("N", "Effective Replication State: $self->{'effrepstate'}, Effective Role: $self->{'effrole'}", 
				116, "$self->{'effrepstate'}", "$self->{'effrole'}");	

	return 0 if ( $self->{'effrole'} eq "primary" );

	return 2 if (($self->{'effrepstate'} eq "Synchronized") &&
			($self->{'effrole'} eq "secondary"));

	return 3 if ($self->{'effrepstate'} eq "Suspended");

	return 4 if ($self->{'effrepstate'} eq "Synchronizing");

	return 5 if ($self->{'effrepstate'} eq "Unsynchronized");

	return 1 if ( $self->{'effrole'} eq "mixed");

	return 6;
}

sub change_role {
	my ($self) = shift;
        my $ldrive = shift;
	my $force = shift;

	my $script_cmd = {};

	if ( $force ) {
	        $script_cmd->{'cmd'} = "\"set remoteMirror localLogicalDrive [\\\"$ldrive\\\"] role=primary force=TRUE autoResync=enabled;\"";
		$script_cmd->{'logcmd'} = "$self->{'smcli'} set remoteMirror localLogicalDrive [$ldrive] role=primary force=TRUE autoResync=enabled;";
	} else {
	        $script_cmd->{'cmd'} = "\"set remoteMirror localLogicalDrive [\\\"$ldrive\\\"] role=primary autoResync=enabled;\"";
		$script_cmd->{'logcmd'} = "$self->{'smcli'} set remoteMirror localLogicalDrive [$ldrive] role=primary autoResync=enabled;";
	}

        if ($self->execute_smcli($script_cmd)) {
		VCSAG_LOG_MSG("E", "Change role request failed with message:$self->{'smcout'}", 113,"$self->{'smcout'}");
                return 1;
	} 

	return 0;
}


sub failover {
	my ($self) = shift;
	my $force = shift;

	my $i = 0;
	my $ret = 0;

        for($i=0; $i < $self->{'numLDs'}; $i++){
                my $ldrive = $self->{LogicalDriveNames}[$i];
		if ( $self->{LD_ROLE}{$ldrive} eq "secondary" ) {
			$ret = $self->change_role($ldrive, $force);
		}

		last if ( $ret );
	}

	return $ret;
}


sub execute_smcli {
	my $self = shift;
	my ($cmdargs) = shift;
	my $smcmd = $self->{'smcli'} . $cmdargs->{'cmd'} ; 
	my $logsmcmd = $cmdargs->{'logcmd'};

	VCSAG_LOG_MSG("N", "Executing smcli command: $logsmcmd", 115, "$logsmcmd");
	
	$self->{'smcout'} = `$smcmd`;

	my $ret = $?;
	if ($ret) {
		VCSAG_LOG_MSG("E", "Execution error of smcli command returned: $self->{'smcout'}", 117, "$self->{'smcout'}");

	}

	return $ret;
}


sub do_command {
        my ($cmd) = @_;
        my $output = `$cmd 2>&1`;
        my $status = $? >> 8;

        return ($status, $output);
}


#
# PingTest - ping host to find out if it is alive
# returns
#  0 if alive
#  1 otherwise
#
sub PingTest {
        my ($host) = @_;
        my $cmd;
	my $SYSTEM;

	$SYSTEM = split('\n', `uname -s`);

	if ($SYSTEM eq "AIX")
	{
        	$cmd="ping -c 1 $host";
	}
	elsif ($SYSTEM eq "HP-UX")
	{
        	$cmd="ping $host -n 1";
	}
	elsif ($SYSTEM eq "Linux")
	{
        	$cmd="ping -c 1 $host";
	}
	elsif ($SYSTEM eq "SunOS")
	{
        	$cmd="ping $host 1";
	}
        my ($status, $output) = do_command($cmd);
        return 1 if ($output =~ /no answer from/);
        return 1 if ($output =~ /unknown host/);
        return 1 if ($output =~ /could not find host/);
        return 1 if ($output =~ /Request timed out/);
	return 1 if ($status);
        return 0;
}

# VMware ESX specific extensions to library
# =========================================
#
sub get_con_str {
        my ($vprovider, $host, $username, $password, $keystore, $code);
        open(CRED, "<$cred_file" || die "Cannot open $cred_file for reading: $!");
        my @lines = <CRED>;
        close (CRED);
        $username = "root";
        my $localhost = `uname -n`;
        chomp($localhost);
        my @vcshostname = split('\s+', `$ag_i18n_inc::HASYS -nodeid`);
        my $vcshost = $vcshostname[5];
        foreach (@lines) {
                if($_ =~ /$vcshost/) {
                        ($host, $code, $password, $keystore) = split('\s+');
                        last;
                }
        }

        #for backwards compatibility with VCS 5.1 GA
        if ($code eq "0") {
                $username = "root";
        } elsif ($code eq "1") {
                $username = "vcs";
	} else {
                $username = $code;
        }

        $vprovider = "1";
        my $dpassword=`$VCSHOME/sbin/vcsdecrypt $password`;
        $host = "https://" . $localhost . "/sdk";
        my @ops = ($vprovider, $host, $username, $dpassword, $keystore);
        my $con_str= join(' ', @ops);
        return $con_str;
}

sub check_vmware_api_err {
        my $localhost = `uname -n`;
        chomp($localhost);
        my $op = $_[0];
        my $errcode = $_[1];
        my $vmtask = $_[2];
        if($errcode == 1) {
                VCSAG_LOG_MSG("E", "VI3 API could not login to $localhost\n", 58, $localhost);
                VCSAG_LOG_MSG("E", "It could either be a network config issue (DNS failure) or invalid localhost credentials file", 59, $localhost);
        } elsif ($errcode == 2) {
                VCSAG_LOG_MSG("E", "$vmtask failed: $localhost\n", 57, $localhost);
        }
}

# rescan hba and vmfs for new datastores
sub exec_vmware_api_cmd {
        my $operation = shift;
        my $localhost = `uname -n`;
        chomp($localhost);
        my $con_str = get_con_str();
        my @ops = ($con_str, $operation, $localhost);
        my $operstr = join(' ', @ops);
        my $op = `$VMWARE_API_CMD $operstr`;
        my $retcode = $? >> 8;
        chomp($op);
        check_vmware_api_err($localhost, $op, $retcode, $operation);
        return 0;
}


sub vmware_online {
        my($self) = shift;

        my $cmd = "";

        exec_vmware_api_cmd("RefreshHostStorage");
	# Make sure that /proc/vmware/config/LVM/EnableResignature and
	# /proc/vmware/config/LVM/DisallowSnapshotLun both are set to 0
	$cmd = "esxcfg-advcfg -s 0 /LVM/EnableResignature";
	system ($cmd);
	$cmd = "esxcfg-advcfg -s 0 /LVM/DisallowSnapshotLun";
	system ($cmd);

	VCSAG_SET_RES_EP_TIMEOUT(180);
	exec_vmware_api_cmd("RescanAllHba");

        return 0;
}

1;
